# -*- coding: utf-8; -*-

import os
import datetime
import decimal
from mock import patch
from unittest import TestCase

import sqlalchemy as sa
from sqlalchemy import orm

from rattail import app as mod
from rattail.config import make_config
from rattail.exceptions import ConfigurationError
from rattail.core import Object
from rattail.db import Session
from rattail.autocomplete import Autocompleter
from rattail.batch import BatchHandler
from rattail.bouncer import BounceHandler
from rattail.importing import ImportHandler
from rattail.importing.rattail import FromRattailToRattailImport
from rattail.gpc import GPC


class TestAppHandler(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.handler = self.make_handler()

    def make_config(self):
        return make_config([], extend=False)

    def make_handler(self):
        return mod.AppHandler(self.config)

    def test_get_title(self):

        # default for unconfigured title
        self.assertEqual(self.handler.get_title(), "Rattail")

        # unless default is provided
        self.assertEqual(self.handler.get_title(default="Foo"), "Foo")

        # or title can be configured
        self.config.setdefault('rattail', 'app_title', 'Bar')
        self.assertEqual(self.handler.get_title(), "Bar")
        self.assertEqual(self.handler.get_title(default="Foo"), "Bar")

    def test_get_timezone(self):

        # unconfigured zone causes error
        self.assertRaises(ConfigurationError, self.handler.get_timezone)

        # or one can be configured
        self.config.setdefault('rattail', 'timezone.default', 'America/Chicago')
        self.assertEqual(str(self.handler.get_timezone()), 'America/Chicago')

        # also can configure alternate zones
        self.assertRaises(ConfigurationError, self.handler.get_timezone, key='other')
        self.config.setdefault('rattail', 'timezone.other', 'America/New_York')
        self.assertEqual(str(self.handler.get_timezone(key='other')), 'America/New_York')

    def test_localtime(self):

        # must define timezone first
        self.config.setdefault('rattail', 'timezone.default', 'America/Chicago')

        # just confirm the method works on a basic level; the
        # underlying function is tested elsewhere
        now = self.handler.localtime()
        self.assertIsNotNone(now)

    def test_make_utc(self):

        # just confirm the method works on a basic level; the
        # underlying function is tested elsewhere
        now = self.handler.make_utc()
        self.assertIsNotNone(now)

    def test_load_object(self):

        # just confirm the method works on a basic level; the
        # underlying function is tested elsewhere
        cls = self.handler.load_object('rattail.core:Object')
        self.assertIs(cls, Object)

    def test_get_active_stores(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        # no stores by default
        stores = self.handler.get_active_stores(session)
        self.assertEqual(len(stores), 0)

        # add a basic store
        store001 = model.Store(id='001')
        session.add(store001)
        session.flush()
        session.refresh(store001)
        self.assertIsNone(store001.archived)

        # that one store should be returned
        stores = self.handler.get_active_stores(session)
        self.assertEqual(len(stores), 1)
        self.assertIs(stores[0], store001)

        # archive first store; add another
        store001.archived = True
        store002 = model.Store(id='002')
        session.add(store002)
        session.flush()
        
        # now only store 002 should be returned
        stores = self.handler.get_active_stores(session)
        self.assertEqual(len(stores), 1)
        self.assertIs(stores[0], store002)

        session.rollback()
        session.close()

    def test_get_autocompleter(self):

        # built-in autocompleter should be got okay
        from rattail.autocomplete.products import ProductAutocompleter
        autocompleter = self.handler.get_autocompleter('products')
        self.assertIsInstance(autocompleter, ProductAutocompleter)

        # now let's invent one, but first make sure it is not yet valid
        self.assertRaises(ValueError, self.handler.get_autocompleter, 'foobars')

        # okay now configure it and then it should be got okay
        self.config.setdefault('rattail', 'autocomplete.foobars',
                               'tests.test_app:FooBarAutocompleter')
        autocompleter = self.handler.get_autocompleter('foobars')
        self.assertIsInstance(autocompleter, FooBarAutocompleter)

    def test_get_auth_handler(self):
        
        # first call gets the default handler
        auth01 = self.handler.get_auth_handler()
        self.assertIsNotNone(auth01)

        # second call gets the same handler instance
        auth02 = self.handler.get_auth_handler()
        self.assertIs(auth02, auth01)

    def test_get_batch_handler(self):

        # unknown batch type raises error by default
        self.assertRaises(ValueError, self.handler.get_batch_handler, 'foobar')

        # or returns None if error is suppressed
        bhandler = self.handler.get_batch_handler('foobar', error=False)
        self.assertIsNone(bhandler)

        # but we can provide our own spec
        bhandler = self.handler.get_batch_handler(
            'foobar', default='tests.test_app:FooBarBatchHandler')
        self.assertIsInstance(bhandler, FooBarBatchHandler)

        # we also can configure our handler
        self.config.setdefault('rattail.batch', 'foobar.handler',
                               'tests.test_app:FooBarBatchHandler')
        bhandler = self.handler.get_batch_handler('foobar')
        self.assertIsInstance(bhandler, FooBarBatchHandler)

        # for some reason (?) the "importer" batch handler is special
        # and can be returned with no config
        from rattail.batch.importer import ImporterBatchHandler
        bhandler = self.handler.get_batch_handler('importer')
        self.assertIsInstance(bhandler, ImporterBatchHandler)

    def test_get_board_handler(self):

        # first call gets the default handler
        board01 = self.handler.get_board_handler()
        self.assertIsNotNone(board01)

        # second call gets the same handler instance
        board02 = self.handler.get_board_handler()
        self.assertIs(board02, board01)

    def test_get_bounce_handler(self):

        # unknown type raises error by default
        self.assertRaises(ValueError, self.handler.get_bounce_handler, 'foobar')

        # but we can configure our own too
        self.config.setdefault('rattail.bouncer', 'foobar.handler',
                               'tests.test_app:FooBarBounceHandler')
        bhandler = self.handler.get_bounce_handler('foobar')
        self.assertIsInstance(bhandler, FooBarBounceHandler)

        # default handler is special and works out of the box
        bhandler = self.handler.get_bounce_handler('default')
        self.assertIsInstance(bhandler, BounceHandler)

    def test_get_clientele_handler(self):

        # first call gets the default handler
        client01 = self.handler.get_clientele_handler()
        self.assertIsNotNone(client01)

        # second call gets the same handler instance
        client02 = self.handler.get_clientele_handler()
        self.assertIs(client02, client01)

    def test_get_custorder_handler(self):

        # first call gets the default handler
        custorder01 = self.handler.get_custorder_handler()
        self.assertIsNotNone(custorder01)

        # second call gets the same handler instance
        custorder02 = self.handler.get_custorder_handler()
        self.assertIs(custorder02, custorder01)

    def test_get_employment_handler(self):

        # first call gets the default handler
        employ01 = self.handler.get_employment_handler()
        self.assertIsNotNone(employ01)

        # second call gets the same handler instance
        employ02 = self.handler.get_employment_handler()
        self.assertIs(employ02, employ01)

    def test_get_feature_handler(self):

        # first call gets the default handler
        feature01 = self.handler.get_feature_handler()
        self.assertIsNotNone(feature01)

        # second call gets the same handler instance
        feature02 = self.handler.get_feature_handler()
        self.assertIs(feature02, feature01)

    def test_get_email_handler(self):

        # first call gets the default handler
        email01 = self.handler.get_email_handler()
        self.assertIsNotNone(email01)

        # second call gets the same handler instance
        email02 = self.handler.get_email_handler()
        self.assertIs(email02, email01)

    def test_get_all_import_handlers(self):

        # several default handlers exist, but not our custom handler
        Handlers = self.handler.get_all_import_handlers()
        self.assertTrue(Handlers)
        self.assertNotIn(FromFooToBar, Handlers)

        # and by default there are no errors to be raised
        Handlers = self.handler.get_all_import_handlers(ignore_errors=False)
        self.assertTrue(Handlers)

        # and just to make sure sorting "works" (no error)
        Handlers = self.handler.get_all_import_handlers(sort=True)
        self.assertTrue(Handlers)

        # finally let's configure a custom handler, and be sure it
        # comes back in the result.  note that we must "override" a
        # default importer here, cannot register a new type without
        # creating an entry point
        self.config.setdefault('rattail.importing',
                               'to_rattail.from_rattail.import.handler',
                               'tests.test_app:FromFooToBar')
        Handlers = self.handler.get_all_import_handlers()
        self.assertTrue(Handlers)
        self.assertIn(FromFooToBar, Handlers)

    def test_get_designated_import_handlers(self):

        # several default handlers exist, but not our custom handler
        handlers = self.handler.get_designated_import_handlers()
        self.assertTrue(handlers)
        self.assertFalse(any([isinstance(h, FromFooToBar)
                              for h in handlers]))
        self.assertTrue(any([isinstance(h, FromRattailToRattailImport)
                             for h in handlers]))

        # we can override a default with custom handler
        self.config.setdefault('rattail.importing',
                               'to_rattail.from_rattail.import.handler',
                               'tests.test_app:FromFooToBar')
        handlers = self.handler.get_designated_import_handlers()
        self.assertTrue(any([isinstance(h, FromFooToBar)
                             for h in handlers]))
        self.assertFalse(any([isinstance(h, FromRattailToRattailImport)
                              for h in handlers]))

        # but then original default is included with alternates
        handlers = self.handler.get_designated_import_handlers(with_alternates=True)
        matches = [h for h in handlers
                   if isinstance(h, FromFooToBar)]
        self.assertEqual(len(matches), 1)
        handler = matches[0]
        self.assertEqual(len(handler.alternate_handlers), 1)
        alternate = handler.alternate_handlers[0]
        self.assertIs(alternate, FromRattailToRattailImport)

    def test_get_import_handler(self):

        # make sure a basic fetch works
        handler = self.handler.get_import_handler('to_rattail.from_rattail.import')
        self.assertIsInstance(handler, FromRattailToRattailImport)

        # and make sure custom override works
        self.config.setdefault('rattail.importing',
                               'to_rattail.from_rattail.import.handler',
                               'tests.test_app:FromFooToBar')
        handler = self.handler.get_import_handler('to_rattail.from_rattail.import')

        # unknown importer cannot be found
        handler = self.handler.get_import_handler('this_should_not_work')
        self.assertIsNone(handler)

        # and if we require it, error will raise
        self.assertRaises(ValueError, self.handler.get_import_handler,
                          'this_should_not_work', require=True)

    def test_get_designated_import_handler_spec(self):
        
        # fetch of unknown key returns none
        spec = self.handler.get_designated_import_handler_spec('test01')
        self.assertIsNone(spec)

        # unless we require it, in which case, error
        self.assertRaises(ValueError, self.handler.get_designated_import_handler_spec,
                          'test01', require=True)

        # we configure one for whatever key we like
        self.config.setdefault('rattail.importing',
                               'test02.handler',
                               'tests.test_app:FromFooToBar')
        spec = self.handler.get_designated_import_handler_spec('test02')
        self.assertEqual(spec, 'tests.test_app:FromFooToBar')

        # we can also define a "default" designated handler
        self.config.setdefault('rattail.importing',
                               'test03.default_handler',
                               'tests.test_app:FromFooToBar')
        spec = self.handler.get_designated_import_handler_spec('test03')
        self.assertEqual(spec, 'tests.test_app:FromFooToBar')

        # we can also designate handler w/ legacy config
        # TODO: this should be removed at some point, surely?
        self.config.setdefault('rattail.importing',
                               'test04.legacy_handler_setting',
                               'rattail.importing, test04.custom_handler')
        self.config.setdefault('rattail.importing',
                               'test04.custom_handler',
                               'tests.test_app:FromFooToBar')
        spec = self.handler.get_designated_import_handler_spec('test04')
        self.assertEqual(spec, 'tests.test_app:FromFooToBar')

    def test_get_label_handler(self):

        # first call gets the default handler
        labels01 = self.handler.get_label_handler()
        self.assertIsNotNone(labels01)

        # second call gets the same handler instance
        labels02 = self.handler.get_label_handler()
        self.assertIs(labels01, labels01)

    def test_get_membership_handler(self):

        # first call gets the default handler
        membership01 = self.handler.get_membership_handler()
        self.assertIsNotNone(membership01)

        # second call gets the same handler instance
        membership02 = self.handler.get_membership_handler()
        self.assertIs(membership02, membership01)

    def test_get_people_handler(self):

        # first call gets the default handler
        people01 = self.handler.get_people_handler()
        self.assertIsNotNone(people01)

        # second call gets the same handler instance
        people02 = self.handler.get_people_handler()
        self.assertIs(people02, people01)

    def test_get_products_handler(self):

        # first call gets the default handler
        products01 = self.handler.get_products_handler()
        self.assertIsNotNone(products01)

        # second call gets the same handler instance
        products02 = self.handler.get_products_handler()
        self.assertIs(products02, products01)

    def test_get_report_handler(self):

        # first call gets the default handler
        report01 = self.handler.get_report_handler()
        self.assertIsNotNone(report01)

        # second call gets the same handler instance
        report02 = self.handler.get_report_handler()
        self.assertIs(report02, report01)

    def test_get_problem_report_handler(self):

        # first call gets the default handler
        problems01 = self.handler.get_problem_report_handler()
        self.assertIsNotNone(problems01)

        # second call gets the same handler instance
        problems02 = self.handler.get_problem_report_handler()
        self.assertIs(problems02, problems01)

    def test_get_trainwreck_handler(self):

        # first call gets the default handler
        trainwreck01 = self.handler.get_trainwreck_handler()
        self.assertIsNotNone(trainwreck01)

        # second call gets the same handler instance
        trainwreck02 = self.handler.get_trainwreck_handler()
        self.assertIs(trainwreck02, trainwreck01)

    def test_get_vendor_handler(self):

        # first call gets the default handler
        vendor01 = self.handler.get_vendor_handler()
        self.assertIsNotNone(vendor01)

        # second call gets the same handler instance
        vendor02 = self.handler.get_vendor_handler()
        self.assertIs(vendor02, vendor01)

    def test_progress_loop(self):
        from rattail.progress import ProgressBase

        class NullProgress(ProgressBase):
            pass

        result = []

        def inspect(obj, i):
            result.append(obj)

        # this is just a basic test to get coverage..
        self.handler.progress_loop(inspect, range(5), NullProgress)
        self.assertEqual(result, list(range(5)))

    def test_make_object(self):

        # basic test
        obj = self.handler.make_object()
        self.assertIsNotNone(obj)

        # make sure attr is set
        obj = self.handler.make_object(answer=42)
        self.assertEqual(obj.answer, 42)

    def test_make_uuid(self):
        uuid = self.handler.make_uuid()
        self.assertIsInstance(uuid, str)
        self.assertEqual(len(uuid), 32)

    def test_get_session(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        # giving an unrelated object raises error
        person = Object()
        self.assertRaises(orm.exc.UnmappedInstanceError,
                          self.handler.get_session, person)

        # a related object still may not be in a session
        person = model.Person()
        result = self.handler.get_session(person)
        self.assertIsNone(result)

        # okay then let's add to session, then should work
        session.add(person)
        result = self.handler.get_session(person)
        self.assertIs(result, session)

        session.rollback()
        session.close()

    def test_make_session(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        
        # default behavior should "work" albeit with no engine bound,
        # and no continuum user set
        session = self.handler.make_session()
        self.assertIsNotNone(session)
        self.assertIsNone(session.bind)
        self.assertIsNone(session.continuum_user)

        # okay then let's create one with engine bound, and add a user
        session = self.handler.make_session(bind=engine)
        user = model.User(username='ferdinand')
        session.add(user)
        session.commit()

        # now we can make a session with that user bound
        session = self.handler.make_session(bind=engine, user='ferdinand')
        self.assertEqual(session.continuum_user.username, 'ferdinand')

        # okay add another user, configure it as default, then confirm
        user = model.User(username='beaufort')        
        session.add(user)
        session.commit()
        self.config.setdefault('rattail', 'runas.default', 'beaufort')
        session = self.handler.make_session(bind=engine)
        self.assertEqual(session.continuum_user.username, 'beaufort')

    def test_cache_model(self):
        engine = sa.create_engine('sqlite://')
        model = self.config.get_model()
        model.Base.metadata.create_all(bind=engine)
        session = Session(bind=engine)

        store001 = model.Store(id='001')
        session.add(store001)
        store002 = model.Store(id='002')
        session.add(store002)
        session.commit()

        # just do a basic cache to prove the concept
        stores = self.handler.cache_model(session, model.Store, key='id')
        self.assertEqual(len(stores), 2)
        self.assertIn('001', stores)
        self.assertIn('002', stores)
        self.assertIs(stores['001'], store001)
        self.assertIs(stores['002'], store002)

    def test_make_temp_dir(self):

        # things work with no args
        path = self.handler.make_temp_dir()
        self.assertTrue(os.path.exists(path))
        os.rmdir(path)

        # we can specify an alternate parent dir (in this case also temp)
        parent = self.handler.make_temp_dir()
        child = self.handler.make_temp_dir(dir=parent)
        self.assertTrue(os.path.exists(child))
        self.assertEqual(os.path.dirname(child), parent)
        os.rmdir(child)
        os.rmdir(parent)

        # also can configure the workdir, to be used as (indirect) parent
        workdir = self.handler.make_temp_dir()
        self.config.setdefault('rattail', 'workdir', workdir)
        child = self.handler.make_temp_dir()
        parent = os.path.dirname(child)
        self.assertEqual(os.path.dirname(parent), workdir)
        os.rmdir(child)
        os.rmdir(parent)
        os.rmdir(workdir)

    def test_make_temp_file(self):

        # things work with no args
        path = self.handler.make_temp_file()
        self.assertTrue(os.path.exists(path))
        os.remove(path)

        # we can specify an alternate parent dir (in this case also temp)
        parent = self.handler.make_temp_dir()
        path = self.handler.make_temp_file(dir=parent)
        self.assertTrue(os.path.exists(path))
        self.assertEqual(os.path.dirname(path), parent)
        os.remove(path)
        os.rmdir(parent)

        # also can configure the workdir, to be used as (indirect) parent
        workdir = self.handler.make_temp_dir()
        self.config.setdefault('rattail', 'workdir', workdir)
        path = self.handler.make_temp_file()
        self.assertTrue(os.path.exists(path))
        parent = os.path.dirname(path)
        self.assertEqual(os.path.dirname(parent), workdir)
        os.remove(path)
        os.rmdir(parent)
        os.rmdir(workdir)

    def normalize_phone_number(self):

        # pre-normalized value is unchanged
        number = '8885551234'
        result = self.handler.normalize_phone_number(number)
        self.assertEqual(result, number)

        # now a basic real-world example
        number = '(888) 555-1234'
        result = self.handler.normalize_phone_number(number)
        self.assertEqual(result, '8885551234')

        # and another for good measure
        number = '888.555.1234'
        result = self.handler.normalize_phone_number(number)
        self.assertEqual(result, '8885551234')

    def test_phone_number_is_invalid(self):
        
        # basic real-world example
        self.assertIsNone(self.handler.phone_number_is_invalid(
            '(888) 555-1234'))

        # and another for good measure
        self.assertIsNone(self.handler.phone_number_is_invalid(
            '888.555.1234'))

        # 10 digits are required, so 9 or 11 digits should fail
        self.assertEqual(self.handler.phone_number_is_invalid('123456789'),
                         "Phone number must have 10 digits")
        self.assertEqual(self.handler.phone_number_is_invalid('12345678901'),
                         "Phone number must have 10 digits")

    def test_format_phone_number(self):
        
        # basic real-world example
        result = self.handler.format_phone_number('8885551234')
        self.assertEqual(result, '(888) 555-1234')

        # garbage in garbage out
        result = self.handler.format_phone_number('garbage')
        self.assertEqual(result, 'garbage')

    def test_make_gpc(self):

        # basic real-world example
        result = self.handler.make_gpc('074305001321')
        self.assertIsInstance(result, GPC)
        self.assertEqual(str(result), '00074305001321')

        # and let it calculate check digit
        result = self.handler.make_gpc('7430500132', calc_check_digit='upc')
        self.assertIsInstance(result, GPC)
        self.assertEqual(str(result), '00074305001321')

    def test_render_gpc(self):

        # basic real-world example
        gpc = GPC('00074305001321')
        result = self.handler.render_gpc(gpc)
        self.assertEqual(result, '0007430500132-1')

    def test_render_currency(self):
        
        # basic decimal example
        value = decimal.Decimal('42.00')
        self.assertEqual(self.handler.render_currency(value), '$42.00')

        # basic float example
        value = 42.00
        self.assertEqual(self.handler.render_currency(value), '$42.00')

        # decimal places will be rounded
        value = decimal.Decimal('42.12345')
        self.assertEqual(self.handler.render_currency(value), '$42.12')

        # but we can declare the scale
        value = decimal.Decimal('42.12345')
        self.assertEqual(self.handler.render_currency(value, scale=4), '$42.1234')

        # negative numbers get parens
        value = decimal.Decimal('-42.42')
        self.assertEqual(self.handler.render_currency(value), '($42.42)')

    def test_render_quantity(self):

        # integer decimals become integers
        value = decimal.Decimal('1.000')
        self.assertEqual(self.handler.render_quantity(value), '1')

        # but decimal places are preserved
        value = decimal.Decimal('1.234')
        self.assertEqual(self.handler.render_quantity(value), '1.234')

    def test_render_cases_units(self):
        
        # basic examples, note the singular noun
        self.assertEqual(self.handler.render_cases_units(1, None), '1 case')
        self.assertEqual(self.handler.render_cases_units(None, 1), '1 unit')

        # mix it up a bit
        self.assertEqual(self.handler.render_cases_units(3, 2), '3 cases + 2 units')

        # also note that zero is not hidden
        self.assertEqual(self.handler.render_cases_units(3, 0), '3 cases + 0 units')

    def test_render_date(self):

        # basic example
        date = datetime.date(2021, 12, 31)
        self.assertEqual(self.handler.render_date(date), '2021-12-31')

    def test_render_datetime(self):

        # basic example
        dt = datetime.datetime(2021, 12, 31, 8, 30)
        self.assertEqual(self.handler.render_datetime(dt), '2021-12-31 08:30:00 AM')

    @patch('rattail.app.send_email')
    def test_send_email(self, send_email):
        
        # just make sure underlying function is invoked..
        self.handler.send_email('test')
        send_email.assert_called()


class FooBarAutocompleter(Autocompleter):
    autocompleter_key = 'foobars'


class FooBarBatchHandler(BatchHandler):
    pass


class FooBarBounceHandler(BounceHandler):
    pass


class FromFooToBar(ImportHandler):
    host_key = 'rattail'
    local_key = 'rattail'



# -*- coding: utf-8; -*-
"""add tender.kick_drawer

Revision ID: d69dd44f495c
Revises: 9cc5b433fb1d
Create Date: 2023-10-06 20:06:56.557531

"""

# revision identifiers, used by Alembic.
revision = 'd69dd44f495c'
down_revision = '9cc5b433fb1d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tender
    op.add_column('tender', sa.Column('kick_drawer', sa.Boolean(), nullable=True))
    op.add_column('tender', sa.Column('disabled', sa.Boolean(), nullable=True))
    op.add_column('tender_version', sa.Column('kick_drawer', sa.Boolean(), autoincrement=False, nullable=True))
    op.add_column('tender_version', sa.Column('disabled', sa.Boolean(), autoincrement=False, nullable=True))

    # batch_pos
    op.alter_column('batch_pos', 'tax1_total', new_column_name='tax_total')
    op.drop_column('batch_pos', 'tax2_total')

    # batch_pos_tax
    op.create_table('batch_pos_tax',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('tax_uuid', sa.String(length=32), nullable=True),
                    sa.Column('tax_code', sa.String(length=30), nullable=False),
                    sa.Column('tax_rate', sa.Numeric(precision=7, scale=5), nullable=False),
                    sa.Column('tax_total', sa.Numeric(precision=9, scale=2), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], ['batch_pos.uuid'], name='batch_pos_tax_fk_batch'),
                    sa.ForeignKeyConstraint(['tax_uuid'], ['tax.uuid'], name='batch_pos_tax_fk_tax'),
                    sa.PrimaryKeyConstraint('uuid')
                    )

    # batch_pow_row
    op.drop_column('batch_pos_row', 'tax2')
    op.drop_column('batch_pos_row', 'tax1')
    op.add_column('batch_pos_row', sa.Column('tax_code', sa.String(length=30), nullable=True))
    op.add_column('batch_pos_row', sa.Column('tender_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('batch_pos_row_fk_tender', 'batch_pos_row', 'tender', ['tender_uuid'], ['uuid'])


def downgrade():

    # batch_pos_row
    op.drop_constraint('batch_pos_row_fk_tender', 'batch_pos_row', type_='foreignkey')
    op.drop_column('batch_pos_row', 'tender_uuid')
    op.drop_column('batch_pos_row', 'tax_code')
    op.add_column('batch_pos_row', sa.Column('tax1', sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.add_column('batch_pos_row', sa.Column('tax2', sa.BOOLEAN(), autoincrement=False, nullable=True))

    # batch_pos_tax
    op.drop_table('batch_pos_tax')

    # batch_pos
    op.add_column('batch_pos', sa.Column('tax2_total', sa.NUMERIC(precision=9, scale=2), autoincrement=False, nullable=True))
    op.alter_column('batch_pos', 'tax_total', new_column_name='tax1_total')

    # tender
    op.drop_column('tender_version', 'disabled')
    op.drop_column('tender_version', 'kick_drawer')
    op.drop_column('tender', 'disabled')
    op.drop_column('tender', 'kick_drawer')



# -*- coding: utf-8; -*-
"""more columns for pos batch

Revision ID: f30d2225fa49
Revises: 6dd836b9829a
Create Date: 2023-09-26 15:31:45.927210

"""

# revision identifiers, used by Alembic.
revision = 'f30d2225fa49'
down_revision = '6dd836b9829a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tender
    op.create_table('tender',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('code', sa.String(length=10), nullable=True),
                    sa.Column('name', sa.String(length=100), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('tender_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.String(length=10), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('notes', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_tender_version_end_transaction_id'), 'tender_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_tender_version_operation_type'), 'tender_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_tender_version_transaction_id'), 'tender_version', ['transaction_id'], unique=False)

    # batch_pos
    op.drop_column('batch_pos', 'start_time')
    op.add_column('batch_pos', sa.Column('terminal_id', sa.String(length=20), nullable=True))
    op.add_column('batch_pos', sa.Column('cashier_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('batch_pos_fk_cashier', 'batch_pos', 'employee', ['cashier_uuid'], ['uuid'])
    op.add_column('batch_pos', sa.Column('customer_is_member', sa.Boolean(), nullable=True))
    op.add_column('batch_pos', sa.Column('customer_is_employee', sa.Boolean(), nullable=True))
    op.add_column('batch_pos', sa.Column('tender_total', sa.Numeric(precision=9, scale=2), nullable=True))
    # nb. hack to get nullable=False for training_mode
    op.add_column('batch_pos', sa.Column('training_mode', sa.Boolean(), nullable=True))
    batch = sa.sql.table('batch_pos', sa.sql.column('training_mode'))
    op.execute(batch.update().values({'training_mode': False}))
    op.alter_column('batch_pos', 'training_mode', nullable=False, existing_type=sa.Boolean())

    # batch_pos_row
    op.add_column('batch_pos_row', sa.Column('timestamp', sa.DateTime(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('user_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('batch_pos_row_fk_user', 'batch_pos_row', 'user', ['user_uuid'], ['uuid'])
    op.add_column('batch_pos_row', sa.Column('upc', rattail.db.types.GPCType(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('batch_pos_row', sa.Column('brand_name', sa.String(length=100), nullable=True))
    op.alter_column('batch_pos_row', 'description', existing_type=sa.VARCHAR(length=60),
                    type_=sa.String(length=255), existing_nullable=True)
    op.add_column('batch_pos_row', sa.Column('size', sa.String(length=255), nullable=True))
    op.add_column('batch_pos_row', sa.Column('full_description', sa.String(length=255), nullable=True))
    op.add_column('batch_pos_row', sa.Column('department_number', sa.Integer(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('department_name', sa.String(length=30), nullable=True))
    op.add_column('batch_pos_row', sa.Column('subdepartment_number', sa.Integer(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('subdepartment_name', sa.String(length=30), nullable=True))
    op.add_column('batch_pos_row', sa.Column('foodstamp_eligible', sa.Boolean(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('sold_by_weight', sa.Boolean(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('cost', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_pos_row', sa.Column('tax1', sa.Boolean(), nullable=True))
    op.add_column('batch_pos_row', sa.Column('tax2', sa.Boolean(), nullable=True))
    op.drop_column('batch_pos_row', 'tax2_total')
    op.drop_column('batch_pos_row', 'tax1_total')
    op.add_column('batch_pos_row', sa.Column('tender_total', sa.Numeric(precision=9, scale=2), nullable=True))


def downgrade():

    # batch_pos_row
    op.drop_column('batch_pos_row', 'tender_total')
    op.add_column('batch_pos_row', sa.Column('tax1_total', sa.NUMERIC(precision=9, scale=2), autoincrement=False, nullable=True))
    op.add_column('batch_pos_row', sa.Column('tax2_total', sa.NUMERIC(precision=9, scale=2), autoincrement=False, nullable=True))
    op.drop_column('batch_pos_row', 'tax2')
    op.drop_column('batch_pos_row', 'tax1')
    op.drop_column('batch_pos_row', 'cost')
    op.drop_column('batch_pos_row', 'sold_by_weight')
    op.drop_column('batch_pos_row', 'foodstamp_eligible')
    op.drop_column('batch_pos_row', 'subdepartment_name')
    op.drop_column('batch_pos_row', 'subdepartment_number')
    op.drop_column('batch_pos_row', 'department_name')
    op.drop_column('batch_pos_row', 'department_number')
    op.drop_column('batch_pos_row', 'full_description')
    op.drop_column('batch_pos_row', 'size')
    op.alter_column('batch_pos_row', 'description', existing_type=sa.String(length=255),
                    type_=sa.VARCHAR(length=60), existing_nullable=True)
    op.drop_column('batch_pos_row', 'brand_name')
    op.drop_column('batch_pos_row', 'item_id')
    op.drop_column('batch_pos_row', 'upc')
    op.drop_constraint('batch_pos_row_fk_user', 'batch_pos_row', type_='foreignkey')
    op.drop_column('batch_pos_row', 'user_uuid')
    op.drop_column('batch_pos_row', 'timestamp')

    # batch_pos
    op.drop_column('batch_pos', 'customer_is_employee')
    op.drop_column('batch_pos', 'customer_is_member')
    op.drop_constraint('batch_pos_fk_cashier', 'batch_pos', type_='foreignkey')
    op.drop_column('batch_pos', 'cashier_uuid')
    op.drop_column('batch_pos', 'training_mode')
    op.drop_column('batch_pos', 'tender_total')
    op.drop_column('batch_pos', 'terminal_id')
    op.add_column('batch_pos', sa.Column('start_time', sa.DateTime(), nullable=True))

    # tender
    op.drop_index(op.f('ix_tender_version_transaction_id'), table_name='tender_version')
    op.drop_index(op.f('ix_tender_version_operation_type'), table_name='tender_version')
    op.drop_index(op.f('ix_tender_version_end_transaction_id'), table_name='tender_version')
    op.drop_table('tender_version')
    op.drop_table('tender')



# -*- coding: utf-8 -*-
"""remove legacy batches

Revision ID: 4d2c5bdfecff
Revises: ea9718f6e6c6
Create Date: 2017-05-13 15:20:08.784955

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '4d2c5bdfecff'
down_revision = 'ea9718f6e6c6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


permission = sa.sql.table('permission', sa.sql.column('permission'))


def rename_perm(oldname, newname):
    op.execute(permission.update().where(permission.c.permission == oldname).values({'permission': newname}))


def upgrade():
    rename_perm('batches.create', 'products.make_batch')

    # batch*
    op.drop_table('batch_column')
    op.drop_table('batch')


def downgrade():
    rename_perm('products.make_batch', 'batches.create')

    # batch
    op.create_table('batch',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('provider', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('id', sa.VARCHAR(length=8), autoincrement=False, nullable=True),
                    sa.Column('source', sa.VARCHAR(length=6), autoincrement=False, nullable=True),
                    sa.Column('destination', sa.VARCHAR(length=6), autoincrement=False, nullable=True),
                    sa.Column('action_type', sa.VARCHAR(length=6), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('rowcount', sa.INTEGER(), autoincrement=False, nullable=True),
                    sa.Column('executed', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
                    sa.Column('purge', sa.DATE(), autoincrement=False, nullable=True),
                    sa.PrimaryKeyConstraint('uuid', name=u'batches_pkey')
    )

    # batch_column
    op.create_table('batch_column',
                    sa.Column('uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=False),
                    sa.Column('batch_uuid', sa.VARCHAR(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.INTEGER(), autoincrement=False, nullable=False),
                    sa.Column('name', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('sil_name', sa.VARCHAR(length=10), autoincrement=False, nullable=True),
                    sa.Column('data_type', sa.VARCHAR(length=15), autoincrement=False, nullable=True),
                    sa.Column('description', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('visible', sa.BOOLEAN(), autoincrement=False, nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'batch.uuid'], name=u'batch_column_fk_batch'),
                    sa.PrimaryKeyConstraint('uuid', name=u'batch_columns_pkey')
    )



# -*- coding: utf-8; -*-
"""add batch_pos.fs_total

Revision ID: 41a05b420280
Revises: e112f75359b3
Create Date: 2023-10-19 16:40:11.395986

"""

# revision identifiers, used by Alembic.
revision = '41a05b420280'
down_revision = 'e112f75359b3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tender
    op.add_column('tender', sa.Column('is_foodstamp', sa.Boolean(), nullable=True))
    op.add_column('tender_version', sa.Column('is_foodstamp', sa.Boolean(), autoincrement=False, nullable=True))

    # batch_pos
    op.add_column('batch_pos', sa.Column('fs_total', sa.Numeric(precision=9, scale=2), nullable=True))
    op.add_column('batch_pos', sa.Column('fs_tender_total', sa.Numeric(precision=9, scale=2), nullable=True))


def downgrade():

    # batch_pos
    op.drop_column('batch_pos', 'fs_tender_total')
    op.drop_column('batch_pos', 'fs_total')

    # tender
    op.drop_column('tender_version', 'is_foodstamp')
    op.drop_column('tender', 'is_foodstamp')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Pricing Utilities
"""

from __future__ import unicode_literals, absolute_import

import decimal
import warnings


def gross_margin(price, cost, percentage=False):
    """
    Calculate and return a gross margin percentage based on ``price`` and
    ``cost``.

    Please note, that for historical reasons, the default behavior is to return
    the margin as a decimal value from 0.0 through 100.0 (or beyond, perhaps).

    However the "industry standard" seems to be to use a decimal value between
    0.000 and 1.000 instead.  Specify ``percentage=True`` for this behavior.

    If ``price`` is empty (or zero), returns ``None``.

    If ``cost`` is empty (or zero), returns ``100`` (or ``1`` if
    ``percentage=True``).
    """
    if not price:
        return None

    if not cost:
        if percentage:
            return 1
        return 100

    margin = (price - cost) / price
    if percentage:
        return margin
    return 100 * margin


def calculate_markup_from_margin(margin, from_decimal=False):
    """
    Calculate the "markup" value corresponding to the given margin.

    This assumes the ``margin`` value is user-friendly, e.g. ``37.5``
    instead of ``0.375`` to represent 37.5%, unless ``from_decimal``
    is true in which case the decimal format is assumed.

    :param margin: Profit margin percentage.

    :param from_decimal: If false (the default), then ``margin``
       should (normally) be between 0 - 100.  But if true, then
       ``margin`` is assumed to be between 0.0 and 1.0 instead.

    :returns: Equivalent cost markup as decimal value (e.g. 1.4).
    """
    if margin is None:
        return
    if margin == 0:
        return 1

    if from_decimal:
        margin *= 100

    return 100 / (100 - margin)


def calculate_markup_from_margin_decimal(margin):
    warnings.warn("calculate_markup_from_margin_decimal() is deprecated; "
                  "please use calculate_markup_from_margin() instead",
                  DeprecationWarning, stacklevel=2)
    return calculate_markup_from_margin(margin, from_decimal=True)


def calculate_markup(margin):
    warnings.warn("calculate_markup() is deprecated; please use "
                  "calculate_markup_from_margin() instead",
                  DeprecationWarning, stacklevel=2)
    return calculate_markup_from_margin(margin, from_decimal=True)


def calculate_variance(oldvalue, newvalue):
    """
    Calculate and return a simple variance percentage, between the
    given old and new values.  For instance::

       calculate_variance(4.00, 6.00)
       # returns 50.0 (50% increase in value)

       calculate_variance(4.00, 3.00)
       # returns -25.0 (25% decrease in value)
    """
    if oldvalue is None or newvalue is None:
        return
    diff = (newvalue or 0) - (oldvalue or 0)
    if not diff:
        return 0
    if not oldvalue:
        return 100
    return diff / oldvalue * 100


def calculate_price_from_margin(cost, margin, places=2):
    """
    Calculate a basic retail price from the given cost amount and
    ideal margin.

    :param cost: Cost amount as decimal.

    :param margin: Ideal / target margin percentage, expressed as
       e.g. 39.0 and *not* 0.39.

    :returns: Price amount as decimal.
    """
    if cost is None or margin is None:
        return

    retail = cost * 100 / (100 - margin)

    # only modify value if caller did *not* specify places=None
    if places is not None:
        if places == 0:
            return int(retail)
        fmt = '{{:0.{}f}}'.format(places)
        return decimal.Decimal(fmt.format(retail))

    return retail



# -*- coding: utf-8 -*-
"""add more newproduct batch row attrs

Revision ID: 3b70d683b0cf
Revises: d39ff41238fa
Create Date: 2022-08-30 09:29:38.260962

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3b70d683b0cf'
down_revision = 'd39ff41238fa'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_newproduct_row
    op.add_column('batch_newproduct_row', sa.Column('unit_size', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('unit_of_measure', sa.String(length=4), nullable=True))
    op.add_column('batch_newproduct_row', sa.Column('unit_of_measure_entry', sa.String(length=20), nullable=True))


def downgrade():

    # batch_newproduct_row
    op.drop_column('batch_newproduct_row', 'unit_of_measure_entry')
    op.drop_column('batch_newproduct_row', 'unit_of_measure')
    op.drop_column('batch_newproduct_row', 'unit_size')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
This subpackage contains logic for the Rattail "datasync" daemon.

For general info see :doc:`rattail-manual:data/sync/index`.
"""

from __future__ import unicode_literals, absolute_import

from .watchers import DataSyncWatcher
from .consumers import (DataSyncConsumer,
                        DataSyncImportConsumer,
                        # TODO: deprecate / remove this
                        NewDataSyncImportConsumer,
                        FromRattailConsumer)
from .rattail import FromRattailToRattailExportConsumer, FromRattailToRattailImportConsumer



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Invoice batches
"""

import decimal
import logging

from rattail.batch import BatchHandler
from rattail.db.model import VendorInvoiceBatch
from rattail.vendors.invoices import require_invoice_parser


log = logging.getLogger(__name__)


class VendorInvoiceHandler(BatchHandler):
    """
    Handler for vendor invoice batches.
    """
    batch_model_class = VendorInvoiceBatch
    po_number_title = "PO Number"

    def make_batch(self, session, progress=None, **kwargs):
        parser_key = kwargs.get('parser_key')
        if parser_key and not kwargs.get('vendor') and not kwargs.get('vendor_uuid'):
            parser = require_invoice_parser(self.config, parser_key)
            vendor_handler = self.app.get_vendor_handler()
            kwargs['vendor'] = vendor_handler.get_vendor(session, parser.vendor_key)
        return super(VendorInvoiceHandler, self).make_batch(session, progress=progress, **kwargs)

    def should_populate(self, batch):
        # all vendor invoices must come from data file
        return True

    def setup_populate(self, batch, progress=None):
        self.vendor = batch.vendor

    def setup_refresh(self, batch, progress=None):
        self.vendor = batch.vendor

    def populate(self, batch, progress=None):
        """
        Pre-fill batch with row data from an input data file, leveraging a
        specific invoice parser.
        """
        assert batch.filename and batch.parser_key
        session = self.app.get_session(batch)
        path = batch.filepath(self.config)
        parser = require_invoice_parser(self.config, batch.parser_key)
        parser.session = session
        vendor_handler = self.app.get_vendor_handler()
        parser.vendor = vendor_handler.get_vendor(session, parser.vendor_key)
        assert batch.vendor is parser.vendor
        try:
            batch.invoice_date = parser.parse_invoice_date(path)
        except:
            log.warning("failed to parse invoice date for file: %s", path,
                        exc_info=True)
        self.setup_populate(batch, progress=progress)

        def append(row, i):
            self.add_row(batch, row)
            if i % 250 == 0:
                session.flush()

        data = list(parser.parse_rows(path))
        assert self.progress_loop(append, data, progress,
                                  message="Adding initial rows to batch")

        # if we have a PO number, attempt to reconcile against a purchase order
        if batch.purchase_order_number:
            purchase = self.get_purchase_order(batch.purchase_order_number)
            if purchase:
                self.cognize_purchase_order(session, batch, purchase, progress=progress)

    def get_purchase_order(self, number):
        """
        Fetch the purchase order object corresponding to the given PO number.
        Custom handlers should override this if aiming to reconcile invoices to
        purchase orders.
        """

    def validate_po_number(self, number):
        """
        This method does nothing by default.  Derived handlers can validate the
        PO number as they like.  Invalid PO numbers should cause ``ValueError``
        to be raised, with text of the reason for validation failure.
        """

    def find_product(self, row):
        """
        Attempt to locate the product for the row, based on UPC etc.
        """
        session = self.app.get_session(row)

        if row.upc:
            product = self.find_product_by_upc(session, row.upc)
            if product:
                return product

        if row.vendor_code:
            product = self.find_product_by_vendor_code(session, row.vendor_code)
            if product:
                return product

    def find_product_by_upc(self, session, upc):
        if hasattr(self, 'products'):
            return self.products['upc'].get(upc)

        handler = self.app.get_products_handler()
        return handler.locate_product_for_gpc(session, upc)

    def find_product_by_vendor_code(self, session, code):
        if hasattr(self, 'products'):
            return self.products['vendor_code'].get(code)

        handler = self.app.get_products_handler()
        return handler.locate_product_for_vendor_code(session, code,
                                                      vendor=self.vendor)

    def refresh_row(self, row):
        """
        Inspect a single row from a invoice, and set its attributes based on
        whether or not the product exists, if we already have a cost record for
        the vendor, if the invoice contains a change etc.  Note that the
        product lookup is done first by UPC and then by vendor item code.
        """
        product = self.find_product(row)
        if not product:
            row.status_code = row.STATUS_NOT_IN_DB
            return

        row.product = product
        row.upc = product.upc
        row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        row.size = product.size

        # fetch case size from product master, if invoice didn't provide
        if not row.case_quantity:
            products_handler = self.app.get_products_handler()
            row.case_quantity = products_handler.get_case_size(product)

        # Calculate case cost if the parser couldn't provide one.
        if not row.case_cost and row.unit_cost and row.case_quantity:
            row.case_cost = row.unit_cost * row.case_quantity

        cost = product.cost_for_vendor(self.vendor)
        if not cost:
            row.status_code = row.STATUS_COST_NOT_IN_DB
        else:

            # calculate unit cost diff, if we have two costs to compare
            if cost.unit_cost is not None and row.unit_cost is not None:
                # note, we must round early here to avoid "false diff"
                row.unit_cost_diff = (row.unit_cost - cost.unit_cost).quantize(
                    decimal.Decimal('0.00001'))

            if not row.case_quantity:
                row.status_code = row.STATUS_NO_CASE_QUANTITY
            elif row.unit_cost_diff:
                row.status_code = row.STATUS_UNIT_COST_DIFFERS
            else:
                row.status_code = row.STATUS_OK

    def cognize_purchase_order(self, session, invoice, purchase, progress=None):
        """
        Cognize the given invoice against the given purchase order object.
        Custom handlers should override this if aiming to reconcile invoices to
        purchase orders.
        """

    def execute(self, batch, **kwargs):
        """
        Execute the vendor invoice batch.  Note that the default handler does
        not perform any actions; a custom handler must be used for anything
        interesting to happen.
        """
        return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for POS transaction batch
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm.collections import attribute_mapped_collection

from rattail.db.model import Base, BatchMixin, BatchRowMixin, uuid_column
from rattail.db.types import GPCType


class POSBatch(BatchMixin, Base):
    """
    Hopefully generic batch used for entering new purchases into the system, etc.?
    """
    batch_key = 'pos'
    __tablename__ = 'batch_pos'
    __batchrow_class__ = 'POSBatchRow'
    model_title = "POS Batch"
    model_title_plural = "POS Batches"

    @declared_attr
    def __table_args__(cls):
        return cls.__batch_table_args__() + (
            sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'],
                                    name='batch_pos_fk_store'),
            sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'],
                                    name='batch_pos_fk_customer'),
            sa.ForeignKeyConstraint(['cashier_uuid'], ['employee.uuid'],
                                    name='batch_pos_fk_cashier'),
        )

    STATUS_OK                           = 1
    STATUS_SUSPENDED                    = 2

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_SUSPENDED                : "suspended",
    }

    store_id = sa.Column(sa.String(length=10), nullable=True, doc="""
    ID of the store where the transaction occurred.
    """)

    store_uuid = sa.Column(sa.String(length=32), nullable=True)
    store = orm.relationship(
        'Store',
        doc="""
        Reference to the store where the transaction ocurred.
        """)

    terminal_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Terminal ID from which the transaction originated, if known.
    """)

    # receipt_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    # Receipt number for the transaction, if known.
    # """)

    # cashier_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    # ID of the cashier who rang the transaction.
    # """)

    # cashier_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    # Name of the cashier who rang the transaction.
    # """)

    # start_time = sa.Column(sa.DateTime(), nullable=True, doc="""
    # UTC timestamp when the transaction began.
    # """)

    # customer_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    # ID of the customer account for the transaction.
    # """)

    # customer_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    # Number of the customer account for the transaction.
    # """)

    # customer_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    # Name of the Customer account for the transaction.
    # """)

    cashier_uuid = sa.Column(sa.String(length=32), nullable=True)
    cashier = orm.relationship(
        'Employee',
        doc="""
        Reference to the employee (cashier) who rang up the transaction.
        """)

    customer_uuid = sa.Column(sa.String(length=32), nullable=True)
    customer = orm.relationship(
        'Customer',
        doc="""
        Reference to the customer account for the transaction.
        """)

    customer_is_member = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the customer was a "member" at time of sale.
    """)

    customer_is_employee = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the customer was an employee at time of sale.
    """)

    # shopper_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    # Number of the shopper account for the transaction, if applicable.
    # """)

    # shopper_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    # Name of the shopper account for the transaction, if applicable.
    # """)

    # shopper_uuid = sa.Column(sa.String(length=32), nullable=True)
    # shopper = orm.relationship(
    #     'CustomerShopper',
    #     doc="""
    #     Reference to the shopper account for the transaction.
    #     """)

    sales_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Sales total for the transaction.
    """)

    fs_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Portion of the sales total which is foodstamp-eligible.
    """)

    tax_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Tax total for the transaction.
    """)

    fs_tender_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Foodstamp tender total for the transaction.
    """)

    tender_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Tender total for the transaction.
    """)

    void = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating if the transaction was voided.
    """)

    training_mode = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating if the transaction was rang in training mode,
    i.e. not real / should not go on the books.
    """)

    def get_balance(self):
        return ((self.sales_total or 0)
                + (self.tax_total or 0)
                + (self.tender_total or 0))

    def get_fs_balance(self):
        return ((self.fs_total or 0)
                + (self.fs_tender_total or 0))


class POSBatchTax(Base):
    """
    A tax total for a POS batch.

    Each row in the batch may be associated with a tax (or not).
    Those which are must be aggregated, to determine overall tax total
    for the batch.  Arbitrary number of taxes may be involved, hence
    we store them in this table.
    """
    __tablename__ = 'batch_pos_tax'
    __table_args__ = (
        sa.ForeignKeyConstraint(['batch_uuid'], ['batch_pos.uuid'],
                                name='batch_pos_tax_fk_batch'),
        sa.ForeignKeyConstraint(['tax_uuid'], ['tax.uuid'],
                                name='batch_pos_tax_fk_tax'),
    )

    uuid = uuid_column()

    batch_uuid = sa.Column(sa.String(length=32), nullable=False)
    batch = orm.relationship(
        POSBatch,
        backref=orm.backref(
            'taxes',
            collection_class=attribute_mapped_collection('tax_code'),
        ))

    tax_uuid = sa.Column(sa.String(length=32), nullable=True)
    tax = orm.relationship('Tax')

    tax_code = sa.Column(sa.String(length=30), nullable=False, doc="""
    Unique "code" for the tax rate.
    """)

    tax_rate = sa.Column(sa.Numeric(precision=7, scale=5), nullable=False, doc="""
    Percentage rate for the tax, e.g. 8.25.
    """)

    tax_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Total for the tax.
    """)


class POSBatchRow(BatchRowMixin, Base):
    """
    Row of data within a POS batch.
    """
    __tablename__ = 'batch_pos_row'
    __batch_class__ = POSBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__batchrow_table_args__() + (
            sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'],
                                    name='batch_pos_row_fk_user'),
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='batch_pos_row_fk_item'),
            sa.ForeignKeyConstraint(['tender_uuid'], ['tender.uuid'],
                                    name='batch_pos_row_fk_tender'),
        )

    STATUS_OK                           = 1

    STATUS = {
        STATUS_OK                       : "ok",
    }

    timestamp = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp when this item was added to transaction.
    """)

    user_uuid = sa.Column(sa.String(length=32), nullable=True)
    user = orm.relationship(
        'User',
        doc="""
        Reference to the user who added this row to the batch.
        """)

    row_type = sa.Column(sa.String(length=32), nullable=True, doc="""
    Type of item represented by this row, e.g. "item" or "return" or
    "tender" etc.

    .. todo::
       need to figure out how to manage/track POSBatchRow.row_type
    """)

    item_entry = sa.Column(sa.String(length=32), nullable=True, doc="""
    Raw/original entry value for the item, if applicable.
    """)

    upc = sa.Column(GPCType(), nullable=True, doc="""
    UPC for the product associated with the row.
    """)

    item_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Generic ID string for the product associated with the row.
    """)

    product_uuid = sa.Column(sa.String(length=32), nullable=True)
    product = orm.relationship(
        'Product',
        doc="""
        Reference to the associated product, if applicable.
        """)

    brand_name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Brand name of the product.
    """)

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Description of the product.
    """)

    size = sa.Column(sa.String(length=255), nullable=True, doc="""
    Size of the product, as string.
    """)

    full_description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Full description for the line item.
    """)

    department_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of the department to which the product belongs.
    """)

    department_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name of the department to which the product belongs.
    """)

    subdepartment_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of the subdepartment to which the product belongs.
    """)

    subdepartment_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name of the subdepartment to which the product belongs.
    """)

    foodstamp_eligible = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the item was eligible for purchase with food
    stamps or equivalent.
    """)

    sold_by_weight = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the item was sold by weight.
    """)

    quantity = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Quantity for the item.
    """)

    cost = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Internal cost for the item sold.

    NOTE: this may need to change at some point, hence the "generic"
    naming so far.  would we need to record multiple kinds of costs?
    """)

    reg_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Regular price for the item.
    """)

    cur_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Current price for the item.
    """)

    cur_price_type = sa.Column(sa.Integer(), nullable=True, doc="""
    Type code for the current price, if applicable.
    """)

    cur_price_start = sa.Column(sa.DateTime(), nullable=True, doc="""
    Start date for current price, if applicable.
    """)

    cur_price_end = sa.Column(sa.DateTime(), nullable=True, doc="""
    End date for current price, if applicable.
    """)

    txn_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Actual price paid for the item.
    """)

    txn_price_adjusted = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the actual price was manually adjusted.
    """)

    sales_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Sales total for the item.
    """)

    tax_code = sa.Column(sa.String(length=30), nullable=True, doc="""
    Unique "code" for the item tax rate, if applicable.
    """)

    tender_total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Tender total for the item.
    """)

    tender_uuid = sa.Column(sa.String(length=32), nullable=True)
    tender = orm.relationship(
        'Tender',
        doc="""
        Reference to the associated tender, if applicable.
        """)

    void = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating the line item was voided.
    """)



# -*- coding: utf-8; -*-
"""
Luigi "backfill" tasks
"""

from __future__ import unicode_literals, absolute_import

import os
import logging

from rattail.config import make_config
from rattail.luigi.tasks import (BackwardBackfillRange, ForwardBackfillRange,
                                 BackfillTask)
from rattail.luigi.logging import WarnSummaryAlways


# nb. BackfillTask must be able to fetch task command
config = make_config()
BackfillTask.config = config

# nb. also we must run from the 'luigi' folder to ensure output
# tracking behaves as expected
os.chdir(os.path.join(config.appdir(), 'luigi'))

# make final luigi summary get logged as a warning
logging.getLogger('luigi-interface').addFilter(WarnSummaryAlways())



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data models for employee work shifts
"""

import sqlalchemy as sa
from sqlalchemy import orm

from .core import Base, uuid_column
from rattail.time import localtime, make_utc


class ShiftMixin(object):
    """
    Mixin for logic common to both scheduled and worked shifts.
    """

    @property
    def length(self):
        if self.start_time and self.end_time:
            return self.end_time - self.start_time

    def get_date(self, config):
        """
        Return the effective date for the shift, according to the given config
        (i.e. timezone).
        """
        time = self.end_time or self.start_time
        assert time
        time = make_utc(time, tzinfo=True)
        return localtime(config, time).date()

    def get_display(self, config):
        """
        Return a simple string for displaying the shift, according to the given
        config (i.e. timezone).
        """
        return '{} - {}'.format(self._format_punch(config, self.start_time),
                                self._format_punch(config, self.end_time))

    def _format_punch(self, config, time):
        if time is None:
            return '??'
        time = localtime(config, make_utc(time, tzinfo=True))
        return time.strftime('%I:%M %p')


class ScheduledShift(Base, ShiftMixin):
    """
    Represents a scheduled shift for an employee.
    """
    __tablename__ = 'employee_shift_scheduled'
    __table_args__ = (
        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_shift_scheduled_fk_employee'),
        sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'], name='employee_shift_scheduled_fk_store'),
    )

    uuid = uuid_column()

    employee_uuid = sa.Column(sa.String(length=32), nullable=False)

    employee = orm.relationship(
        'Employee',
        doc="""
        Reference to the :class:`rattail.db.model.Employee` instance whose
        shift this is.
        """,
        backref=orm.backref(
            'scheduled_shifts',
            cascade_backrefs=False,
            doc="""
            Sequence of :class:`rattail.db.model.ScheduledShift` instances for the
            employee.
            """))

    store_uuid = sa.Column(sa.String(length=32), nullable=True)

    store = orm.relationship(
        'Store',
        doc="""
        Reference to the :class:`rattail.db.model.Store` instance, representing
        the location at which the shift is scheduled, if applicable/known.
        """)

    start_time = sa.Column(sa.DateTime(), nullable=False, doc="""
    Date and time when the shift is scheduled to start.
    """)

    end_time = sa.Column(sa.DateTime(), nullable=False, doc="""
    Date and time when the shift is scheduled to end.
    """)


class WorkedShift(Base, ShiftMixin):
    """
    Represents a shift actually *worked* by an employee.  (Either ``punch_in``
    or ``punch_out`` is generally assumed to be non-null.)
    """
    __tablename__ = 'employee_shift_worked'
    __table_args__ = (
        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_shift_worked_fk_employee'),
        sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'], name='employee_shift_worked_fk_store'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    employee_uuid = sa.Column(sa.String(length=32), nullable=False)

    employee = orm.relationship(
        'Employee',
        doc="""
        Reference to the :class:`rattail.db.model.Employee` instance whose
        shift this is.
        """,
        backref=orm.backref(
            'worked_shifts',
            cascade_backrefs=False,
            doc="""
            Sequence of :class:`rattail.db.model.WorkedShift` instances for the
            employee.
            """))

    store_uuid = sa.Column(sa.String(length=32), nullable=True)

    store = orm.relationship(
        'Store',
        doc="""
        Reference to the :class:`rattail.db.model.Store` instance, representing
        the location at which the shift was worked, if applicable/known.
        """)

    punch_in = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp representing the punch-in time for the shift.
    """)

    punch_out = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp representing the punch-out time for the shift.
    """)

    # TODO: These are needed to make WorkedShift interchangeable with
    # ScheduledShift for certain UI logic etc.  Perhaps should just rename the
    # 'punch' columns at some point..?
    start_time = orm.synonym('punch_in')
    end_time = orm.synonym('punch_out')

    def __str__(self):
        return str(self.employee or '')



# -*- coding: utf-8 -*-
"""add employee_x_department

Revision ID: 57d12767eb53
Revises: 40326eb83e18
Create Date: 2016-01-13 21:53:29.946003

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '57d12767eb53'
down_revision = u'40326eb83e18'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # employee_x_department
    op.create_table('employee_x_department',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), nullable=False),
                    sa.Column('department_uuid', sa.String(length=32), nullable=False),
                    sa.ForeignKeyConstraint(['department_uuid'], [u'department.uuid'], name=u'employee_x_department_fk_department'),
                    sa.ForeignKeyConstraint(['employee_uuid'], [u'employee.uuid'], name=u'employee_x_department_fk_employee'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # employee_x_department
    op.drop_table('employee_x_department')



# -*- coding: utf-8; -*-
"""add CustomerShopper

Revision ID: a4ac7985cfa6
Revises: 2ff00a7cb734
Create Date: 2023-06-06 20:55:28.816311

"""

# revision identifiers, used by Alembic.
revision = 'a4ac7985cfa6'
down_revision = '2ff00a7cb734'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # customer
    op.add_column('customer', sa.Column('account_holder_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('customer_fk_account_holder', 'customer', 'person', ['account_holder_uuid'], ['uuid'])
    op.add_column('customer_version', sa.Column('account_holder_uuid', sa.String(length=32), autoincrement=False, nullable=True))

    # customer_shopper
    op.create_table('customer_shopper',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('customer_uuid', sa.String(length=32), nullable=False),
                    sa.Column('person_uuid', sa.String(length=32), nullable=False),
                    sa.Column('shopper_number', sa.Integer(), nullable=False),
                    sa.Column('active', sa.Boolean(), nullable=True),
                    sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='customer_shopper_fk_customer'),
                    sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='customer_shopper_fk_person'),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('customer_uuid', 'shopper_number', name='customer_shopper_uq_shopper_number')
                    )
    op.create_index('customer_shopper_ix_customer', 'customer_shopper', ['customer_uuid'], unique=False)
    op.create_index('customer_shopper_ix_person', 'customer_shopper', ['person_uuid'], unique=False)
    op.create_table('customer_shopper_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('customer_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('shopper_number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('active', sa.Boolean(), nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_customer_shopper_version_end_transaction_id'), 'customer_shopper_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_shopper_version_operation_type'), 'customer_shopper_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_shopper_version_transaction_id'), 'customer_shopper_version', ['transaction_id'], unique=False)

    # customer_shopper_history
    op.create_table('customer_shopper_history',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('shopper_uuid', sa.String(length=32), nullable=False),
                    sa.Column('start_date', sa.Date(), nullable=True),
                    sa.Column('end_date', sa.Date(), nullable=True),
                    sa.ForeignKeyConstraint(['shopper_uuid'], ['customer_shopper.uuid'], name='customer_shopper_history_fk_shopper'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_index('customer_shopper_history_ix_shopper', 'customer_shopper_history', ['shopper_uuid'], unique=False)
    op.create_table('customer_shopper_history_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('shopper_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('start_date', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('end_date', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_customer_shopper_history_version_end_transaction_id'), 'customer_shopper_history_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_shopper_history_version_operation_type'), 'customer_shopper_history_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_shopper_history_version_transaction_id'), 'customer_shopper_history_version', ['transaction_id'], unique=False)


def downgrade():

    # customer_shopper_history
    op.drop_index(op.f('ix_customer_shopper_history_version_transaction_id'), table_name='customer_shopper_history_version')
    op.drop_index(op.f('ix_customer_shopper_history_version_operation_type'), table_name='customer_shopper_history_version')
    op.drop_index(op.f('ix_customer_shopper_history_version_end_transaction_id'), table_name='customer_shopper_history_version')
    op.drop_table('customer_shopper_history_version')
    op.drop_index('customer_shopper_history_ix_shopper', table_name='customer_shopper_history')
    op.drop_table('customer_shopper_history')

    # customer_shopper
    op.drop_index(op.f('ix_customer_shopper_version_transaction_id'), table_name='customer_shopper_version')
    op.drop_index(op.f('ix_customer_shopper_version_operation_type'), table_name='customer_shopper_version')
    op.drop_index(op.f('ix_customer_shopper_version_end_transaction_id'), table_name='customer_shopper_version')
    op.drop_table('customer_shopper_version')
    op.drop_index('customer_shopper_ix_person', table_name='customer_shopper')
    op.drop_index('customer_shopper_ix_customer', table_name='customer_shopper')
    op.drop_table('customer_shopper')

    # customer
    op.drop_column('customer_version', 'account_holder_uuid')
    op.drop_constraint('customer_fk_account_holder', 'customer', type_='foreignkey')
    op.drop_column('customer', 'account_holder_uuid')



# -*- coding: utf-8 -*-
"""add product_batch_row.item_entry

Revision ID: 0a3dca1cd299
Revises: 42edf57f1ae2
Create Date: 2018-09-22 19:56:36.431827

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '0a3dca1cd299'
down_revision = u'42edf57f1ae2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # misc. batches
    op.add_column('batch_handheld_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('label_batch_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('item_entry', sa.String(length=20), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('item_entry', sa.String(length=20), nullable=True))


def downgrade():

    # misc. batches
    op.drop_column('vendor_invoice_row', 'item_entry')
    op.drop_column('vendor_catalog_row', 'item_entry')
    op.drop_column('purchase_batch_row', 'item_entry')
    op.drop_column('label_batch_row', 'item_entry')
    op.drop_column('batch_pricing_row', 'item_entry')
    op.drop_column('batch_inventory_row', 'item_entry')
    op.drop_column('batch_handheld_row', 'item_entry')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi utilities
"""

import os
import logging
import subprocess
import sys
import warnings
from xmlrpc.client import ProtocolError

import sqlalchemy as sa

from rattail.app import GenericHandler
from rattail.util import shlex_join


log = logging.getLogger(__name__)


class LuigiHandler(GenericHandler):
    """
    Base class and default implementation for Luigi handler.
    """

    def get_supervisor_process_name(self, require=False, **kwargs):
        getter = self.config.require if require else self.config.get
        return getter('rattail.luigi', 'scheduler.supervisor_process_name')

    def restart_supervisor_process(self, name=None, **kwargs):
        if not name:
            name = self.get_supervisor_process_name()

        try:
            proxy = self.app.make_supervisorctl_proxy()
        except:
            log.warning("failed to make supervisorctl proxy", exc_info=True)

        else:
            # we have our proxy, so use that, then return
            try:
                info = proxy.supervisor.getProcessInfo(name)
                if info['state'] != 0:
                    proxy.supervisor.stopProcess(name)
                proxy.supervisor.startProcess(name)
            except ProtocolError as error:
                raise self.app.safe_supervisor_protocol_error(error)
            return

        # no proxy, but we can still try command line
        # TODO: should rename this setting at some point?
        cmd = self.config.get('rattail.luigi', 'scheduler.restart_command')
        if cmd:
            cmd = self.config.parse_list(cmd)
        elif name:
            cmd = ['supervisorctl', 'restart', name]

        log.debug("attempting luigi scheduler restart with command: %s", cmd)

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as error:
            log.warning("failed to restart luigi scheduler; output was:")
            log.warning(error.output)
            raise

    def get_all_overnight_tasks(self, **kwargs):
        tasks = []

        keys = self.config.getlist('rattail.luigi', 'overnight.tasks',
                                   default=[])
        if not keys:
            keys = self.config.getlist('rattail.luigi', 'overnight_tasks',
                                       default=[])
            if keys:
                warnings.warn("setting is deprecated: [rattail.luigi] overnight_tasks; "
                              "please use [rattail.luigi] overnight.tasks instead",
                              DeprecationWarning, stacklevel=2)

        for key in keys:
            if key.startswith('overnight-'):
                key = key[len('overnight-'):]
                warnings.warn("overnight task keys use deprecated 'overnight-' prefix",
                              DeprecationWarning, stacklevel=2)

            lastrun = self.get_overnight_task_setting(key, 'lastrun')
            lastrun = self.app.parse_utctime(lastrun, local=True)
            tasks.append({
                'key': key,
                'description': self.get_overnight_task_setting(key, 'description'),
                'module': self.get_overnight_task_setting(key, 'module'),
                'class_name': self.get_overnight_task_setting(key, 'class_name'),
                'script': self.get_overnight_task_setting(key, 'script'),
                'notes': self.get_overnight_task_setting(key, 'notes'),
                'lastrun': lastrun,
                'last_date': self.get_overnight_task_setting(key, 'last_date',
                                                             typ='date'),
            })
        tasks.sort(key=lambda t: t['description'])
        return tasks

    def get_overnight_task_setting(self, key, name, typ=None, **kwargs):
        getter = self.config.get
        if typ == 'date':
            getter = self.config.getdate
        value = getter('rattail.luigi',
                       'overnight.task.{}.{}'.format(key, name))
        if value is None:
            value = getter('rattail.luigi',
                           'overnight.overnight-{}.{}'.format(key, name))
            if value is not None:
                warnings.warn("[rattail.luigi] overnight.overnight-* settings are deprecated; "
                              "please use [rattail.luigi] overnight.task.* instead",
                              DeprecationWarning, stacklevel=2)
        return value

    def get_overnight_task(self, key, **kwargs):
        if key.startswith('overnight-'):
            key = key[len('overnight-'):]
            warnings.warn("overnight task keys use deprecated 'overnight-' prefix",
                          DeprecationWarning, stacklevel=2)

        for task in self.get_all_overnight_tasks():
            if task['key'] == key:
                return task

    def launch_overnight_task(self, task, date,
                              keep_config=True,
                              email_if_empty=True,
                              email_key=None,
                              wait=True,
                              dry_run=False,
                              **kwargs):
        """
        Launch the given overnight task, to run for the given date.

        :param task: An overnight task info dict, e.g. as obtained
           from :meth:`get_overnight_task()`.

        :param date: Date for which task should run.

        :param keep_config: If true (the default) then the ``rattail
           overnight`` command will be invoked with the same config
           file(s) which are effective in the current/parent process.
           If false, it will be invoked only with ``app/silent.conf``.

        :param email_if_empty: If true (the default), then email will
           be sent when the task command completes, even if it
           produces no output.  If false, then email is sent only if
           the command produces output.

        :param email_key: Optional config key for email settings to be
           used in determining recipients etc.

        :param wait: If true (the default), the task will run
           in-process, and so will begin immediately, but caller must
           wait for it to complete.  If false, the task will be
           scheduled via the ``at`` command, to begin within the next
           minute.  (This lets process control return immediately to
           the caller.)

        :param dry_run: If true, log the final command for the task
           but do not actually run it.
        """
        appdir = self.config.appdir()
        env = {'PYTHONPATH': appdir}
        if keep_config:
            env['RATTAIL_CONFIG_FILES'] = os.pathsep.join(self.config.files_read)
        else:
            env['RATTAIL_CONFIG_FILES'] = os.path.join(appdir, 'silent.conf')

        # build our command, which will vary per caller request.  by
        # default we do not use the shell to run command, but in some
        # cases must (e.g. when invoking `at`)
        shell = False

        # not waiting means schedule via `at`
        if not wait:

            # rattail overnight <task>
            cmd = [
                '{}/bin/rattail'.format(sys.prefix),
                '--no-versioning',
                'overnight', '-k', task['key'],
                '--date={}'.format(date),
                '--wait',
            ]
            if email_key:
                cmd.extend(['--email-key', email_key])
            if not email_if_empty:
                cmd.append('--no-email-if-empty')

            # echo 'rattail overnight <task>' | at now
            cmd = ['echo', shlex_join(cmd)]
            cmd = shlex_join(cmd)
            cmd = "{} | at 'now + 1 minute'".format(cmd)
            shell = True        # must run command via shell

        else:
            # run-n-mail task immediately

            if task['script']:
                # use whatever script is configured
                cmd = self.config.parse_list(task['script'])

            else:
                # invoke luigi directly
                cmd = [
                    '{}/bin/luigi'.format(sys.prefix),
                    '--module', task['module'],
                    task['class_name'],
                    '--date={}'.format(date),
                ]

            # rattail run-n-mail 'task command'
            cmd = [
                os.path.join(sys.prefix, 'bin', 'rattail'),
                '--no-versioning',
                'run-n-mail',
                '--keep-exit-code',
                '--subject', "Overnight for {}: {}".format(date, task['key']),
                shlex_join(cmd),
            ]
            if email_key:
                cmd.extend(['--key', email_key])
            if not email_if_empty:
                cmd.append('--skip-if-empty')

        # log final command
        log.debug("env is: %s", env)
        log.debug("launching command in subprocess: %s", cmd)
        if dry_run:
            log.debug("dry-run mode, so aborting")
            return

        # run command in subprocess
        curdir = os.getcwd()
        try:
            # nb. always chdir to luigi folder, even if not needed
            os.chdir(os.path.join(appdir, 'luigi'))
            subprocess.check_output(cmd, shell=shell, env=env,
                                    stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as error:
            log.warning("command failed with exit code %s!  output was:",
                        error.returncode)
            log.warning(error.output.decode('utf_8'))
            raise
        else:
            if wait and self.config.getbool('rattail.luigi',
                                            'auto_record_last_date',
                                            # TODO: default should be true
                                            default=False):
                self.record_overnight_last_date(task, date)
        finally:
            # nb. always change back to first dir, in case we're being
            # called by some long-running process, e.g. web app
            os.chdir(curdir)

    def record_overnight_last_date(self, task, date, session=None, **kwargs):
        name = 'rattail.luigi.overnight.task.{}.last_date'.format(task['key'])
        with self.app.short_session(session=session, commit=True) as s:
            self.app.save_setting(s, name, str(date))

    def get_all_backfill_tasks(self, **kwargs):
        tasks = []

        keys = self.config.getlist('rattail.luigi', 'backfill.tasks',
                                   default=[])
        if not keys:
            keys = self.config.getlist('rattail.luigi', 'backfill_tasks',
                                       default=[])
            if keys:
                warnings.warn("setting is deprecated: [rattail.luigi] backfill_tasks; "
                              "please use [rattail.luigi] backfill.tasks instead",
                              DeprecationWarning, stacklevel=2)

        for key in keys:
            if key.startswith('backfill-'):
                key = key[len('backfill-'):]
                warnings.warn("backfill task keys use deprecated 'backfill-' prefix",
                              DeprecationWarning, stacklevel=2)

            lastrun = self.get_backfill_task_setting(key, 'lastrun')
            lastrun = self.app.parse_utctime(lastrun, local=True)
            tasks.append({
                'key': key,
                'description': self.get_backfill_task_setting(key, 'description'),
                'script': self.get_backfill_task_setting(key, 'script'),
                'forward': self.get_backfill_task_setting(key, 'forward',
                                                          typ='bool') or False,
                'notes': self.get_backfill_task_setting(key, 'notes'),
                'lastrun': lastrun,
                'last_date': self.get_backfill_task_setting(key, 'last_date',
                                                            typ='date'),
                'target_date': self.get_backfill_task_setting(key, 'target_date',
                                                              typ='date'),
            })
        tasks.sort(key=lambda t: t['description'])
        return tasks

    def get_backfill_task_setting(self, key, name, typ=None, **kwargs):
        getter = self.config.get
        if typ == 'bool':
            getter = self.config.getbool
        elif typ == 'date':
            getter = self.config.getdate
        value = getter('rattail.luigi',
                       'backfill.task.{}.{}'.format(key, name))
        if value is None:
            value = getter('rattail.luigi',
                           'backfill.{}.{}'.format(key, name))
            if value is not None:
                warnings.warn("[rattail.luigi] backfill.* settings are deprecated; "
                              "please use [rattail.luigi] backfill.task.* instead",
                              DeprecationWarning, stacklevel=2)
        return value

    def get_backfill_task(self, key, **kwargs):
        if key.startswith('backfill-'):
            key = key[len('backfill-'):]
            warnings.warn("backfill task keys use deprecated 'backfill-' prefix",
                          DeprecationWarning, stacklevel=2)

        for task in self.get_all_backfill_tasks():
            if task['key'] == key:
                return task

    def launch_backfill_task(self, task, start_date, end_date,
                             keep_config=True,
                             email_if_empty=True,
                             email_key=None,
                             wait=True,
                             dry_run=False,
                             **kwargs):
        """
        Launch the given backfill task, to run for the given date
        range.

        :param task: A backfill task info dict, e.g. as obtained from
           :meth:`get_backfill_task()`.

        :param start_date: Start of date range for which task should
           run.

        :param end_date: End of date range for which task should run.
           This is *inclusive* so the task will be ran for this
           ``end_date`` value, at the end.

        :param keep_config: If true (the default) then the subcommand
           will be invoked with the same config file(s) which are
           effective in the current/parent process.  If false, it will
           be invoked only with ``app/silent.conf``.

        :param email_if_empty: If true (the default), then email will
           be sent when the task command completes, even if it
           produces no output.  If false, then email is sent only if
           the command produces output.

        :param email_key: Optional config key for email settings to be
           used in determining recipients etc.

        :param wait: If true (the default), the task will run
           in-process, and so will begin immediately, but caller must
           wait for it to complete.  If false, the task will be
           scheduled via the ``at`` command, to begin within the next
           minute.  (This lets process control return immediately to
           the caller.)

        :param dry_run: If true, log the final command for the task
           but do not actually run it.
        """
        if not start_date or not end_date:
            raise ValueError("must specify both start_date and end_date")

        if start_date > end_date:
            start_date, end_date = end_date, start_date

        appdir = self.config.appdir()

        env = {
            # TODO: is this ever needed here?
            # 'PYTHONPATH': appdir,
        }
        if keep_config:
            env['RATTAIL_CONFIG_FILES'] = os.pathsep.join(self.config.files_read)
        else:
            env['RATTAIL_CONFIG_FILES'] = os.path.join(appdir, 'silent.conf')

        # build our command, which will vary per caller request.  by
        # default we do not use the shell to run command, but in some
        # cases must (e.g. when invoking `at`)
        shell = False

        # luigi
        luigi = os.path.join(sys.prefix, 'bin', 'luigi')
        cmd = [luigi, '--module=rattail.luigi.backfill_runner',
               '{}BackfillRange'.format(
                   'Forward' if task['forward'] else 'Backward'),
               '--key', task['key'],
               '--start-date={}'.format(start_date),
               '--end-date={}'.format(end_date)]

        # rattail run-n-mail luigi
        last_date = end_date if task['forward'] else start_date
        cmd = [os.path.join(sys.prefix, 'bin', 'rattail'),
               '--no-versioning',
               'run-n-mail',
               '--keep-exit-code',
               '--subject', 'Backfill thru {}: {}'.format(last_date, task['description']),
               shlex_join(cmd)]
        if email_key:
            cmd.extend(['--key', email_key])
        if not email_if_empty:
            cmd.append('--skip-if-empty')

        # not waiting means schedule via `at`
        if not wait:

            # echo 'rattail run-n-mail luigi' | at now
            cmd = ['echo', shlex_join(cmd)]
            cmd = shlex_join(cmd)
            cmd = "{} | at 'now + 1 minute'".format(cmd)
            shell = True        # must run command via shell

        # log final command
        log.debug("env is: %s", env)
        log.debug("launching command in subprocess: %s", cmd)
        if dry_run:
            log.debug("dry-run mode, so aborting")
            return

        # run command in subprocess
        curdir = os.getcwd()
        try:
            # nb. always chdir to luigi folder, even if not needed
            os.chdir(os.path.join(appdir, 'luigi'))
            subprocess.check_output(cmd, shell=shell, env=env,
                                    stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as error:
            log.warning("command failed with exit code %s!  output was:",
                        error.returncode)
            log.warning(error.output.decode('utf_8'))
            raise
        finally:
            # nb. always change back to first dir, in case we're being
            # called by some long-running process, e.g. web app
            os.chdir(curdir)

    def record_backfill_last_date(self, task, date, session=None, **kwargs):
        name = 'rattail.luigi.backfill.task.{}.last_date'.format(task['key'])
        with self.app.short_session(session=session, commit=True) as s:
            self.app.save_setting(s, name, str(date))

    def purge_overnight_settings(self, session):
        model = self.model
        to_delete = session.query(model.Setting)\
                           .filter(sa.or_(
                               model.Setting.name == 'rattail.luigi.overnight.tasks',
                               model.Setting.name.like('rattail.luigi.overnight.task.%.description'),
                               model.Setting.name.like('rattail.luigi.overnight.task.%.notes'),
                               model.Setting.name.like('rattail.luigi.overnight.task.%.module'),
                               model.Setting.name.like('rattail.luigi.overnight.task.%.class_name'),
                               model.Setting.name.like('rattail.luigi.overnight.task.%.script'),
                               # TODO: these should no longer be used
                               model.Setting.name == 'rattail.luigi.overnight_tasks',
                               model.Setting.name.like('rattail.luigi.overnight.%.description'),
                               model.Setting.name.like('rattail.luigi.overnight.%.notes'),
                               model.Setting.name.like('rattail.luigi.overnight.%.script')))\
                           .all()
        for setting in to_delete:
            self.app.delete_setting(session, setting.name)

    def purge_backfill_settings(self, session):
        model = self.model
        to_delete = session.query(model.Setting)\
                           .filter(sa.or_(
                               model.Setting.name == 'rattail.luigi.backfill.tasks',
                               model.Setting.name.like('rattail.luigi.backfill.task.%.description'),
                               model.Setting.name.like('rattail.luigi.backfill.task.%.forward'),
                               model.Setting.name.like('rattail.luigi.backfill.task.%.notes'),
                               model.Setting.name.like('rattail.luigi.backfill.task.%.script'),
                               model.Setting.name.like('rattail.luigi.backfill.task.%.target_date'),
                               # TODO: these should no longer be used
                               model.Setting.name == 'rattail.luigi.backfill_tasks',
                               model.Setting.name.like('rattail.luigi.backfill.%.description'),
                               model.Setting.name.like('rattail.luigi.backfill.%.forward'),
                               model.Setting.name.like('rattail.luigi.backfill.%.notes'),
                               model.Setting.name.like('rattail.luigi.backfill.%.script'),
                               model.Setting.name.like('rattail.luigi.backfill.%.target_date')))\
                           .all()
        for setting in to_delete:
            self.app.delete_setting(session, setting.name)
